from docutils.core import publish_doctree
from docutils import nodes
import sys


def create_node(text):
    """
    Converts a reStructuredText String into a docutils.node
    This node can be included in the document.
    """
    return publish_doctree(text).children


def trim(docstring):
    """
    Removes spaces and extra tabs from a docstring
    Implementation from http://www.python.org/dev/peps/pep-0257/
    """
    if not docstring:
        return ''
    # Convert tabs to spaces (following the normal Python rules)
    # and split into a list of lines:
    lines = docstring.expandtabs().splitlines()
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxint
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxint:
        for line in lines[1:]:
            trimmed.append(line[indent:].rstrip())
    # Strip off trailing and leading blank lines:
    while trimmed and not trimmed[-1]:
        trimmed.pop()
    while trimmed and not trimmed[0]:
        trimmed.pop(0)
    # Return a single string:
    res = '\n'.join(trimmed)
    if not isinstance(res, unicode):
        res = res.decode('utf8')
    return res


def empty_node():
    return nodes.inline(text="")



from docutils import nodes
from docutils.parsers.rst import Directive
from lovely.pyrest.rest import get_services
from importlib import import_module
from lovely.pyrest.sphinx.helpers import create_node, trim
from pyramid import paster
from pyramid.config import Configurator
import json
import venusian


def convert_to_list(value):
    if value is None:
        return []
    return [v.strip() for v in value.split(',')]


class ServiceDirective(Directive):
    """ The Service directive renders all services defined in a module """

    has_content = True
    required_arguments = 0
    optional_arguments = 1

    option_spec = {'services': convert_to_list}

    def run(self):
        # Check if the pyramid config file is set, if souse paster to manage
        # imports
        conf_file = self.pyramid_conf()
        if conf_file:
            paster.get_app(conf_file)
        else:
            # instantiate a pyramid configurator
            config = Configurator()
            module_str = self.arguments[0]
            # import the module
            module = import_module(module_str)
            # check if the module has an `includeme` method and call it
            # because the base route must be added
            if hasattr(module, 'includeme'):
                module.includeme(config)
            config.commit()
            # scan the module for services
            scanner = venusian.Scanner(config=config)
            scanner.scan(module)
        rendered = []
        # fetch all services
        services = get_services()
        # if the services option is set render only the named services
        names = self.options.get('services')
        if names:
            services = [s for s in services if s.baseRouteName in names]
        for service in services:
            service_id = "service_%d" % self.serialno('service')
            rendered.append(ServiceDirective.render(service, service_id))
        return rendered

    def serialno(self, ident):
        """ Renders a serialno for a given term
            Should be overwritten in tests
        """
        settings = self.state.document.settings
        return settings.env.new_serialno(ident)

    def pyramid_conf(self):
        conf = self.state.document.settings.env.config
        return conf['pyramid_conf']

    @staticmethod
    def render(service, service_id):
        service_node = nodes.section(ids=[service_id])
        title = "%s service" % service.baseRouteName
        service_node += nodes.title(text=title)
        if service.description is not None:
            service_node += create_node(trim(service.description))

        for pattern, route_kw, view_kw, func, schema in service.methods:
            method = route_kw.get('request_method', 'GET')
            method_id = "%s_%s" % (service_id, method)
            method_node = nodes.section(ids=[method_id])
            desc = func.__doc__
            title = None
            if desc:
                # use the first line as title for the endpoint
                sp = desc.split('\n', 1)
                title = sp[0].strip()
                if len(sp) > 1:
                    desc = sp[1].strip()
                else:
                    desc = ''
            if not title:
                title = '%s - %s' % (method, pattern)
            method_node += nodes.title(text=title)
            url = "::\n\n    \n    %s - %s\n\n" % (method, pattern)
            method_node += create_node(url)
            # render description from docstring
            if desc:
                method_node += create_node(trim(desc))

            accept = route_kw.get('accept')
            content_type = route_kw.get('content_type')

            # Render Accept Header documentation
            if accept:
                accept_desc = 'Accept: %s' % accept
                method_node += create_node(accept_desc)
            # Render Content-Type Header documentation
            if content_type:
                content_desc = 'Content-Type: %s' % content_type
                method_node += create_node(content_desc)

            # Render Validator
            if schema:
                schema_id = "%s_%s_%s" % (service_id,
                                          method,
                                          'validator')
                node = nodes.section(ids=[schema_id])
                title = nodes.title(text='Validation Schema')
                node += title
                text = json.dumps(schema, indent=4)
                # indent the text block
                text = '\n    '.join([l for l in text.splitlines()])
                text = '::\n\n    ' + text + '\n\n'
                node += create_node(text)

                method_node += node
            service_node += method_node
        return service_node



import os
import re
import ConfigParser
from setuptools import setup, find_packages

#import lovely.pyrest

versionf_content = open("lovely/pyrest/__init__.py").read()
version_rex = r'^__version__ = [\'"]([^\'"]*)[\'"]$'
m = re.search(version_rex, versionf_content, re.M)
if m:
    version = m.group(1)
else:
    raise RuntimeError('Unable to find version string')


def get_versions():
    """picks the versions from version.cfg and returns them as dict"""
    versions_cfg = os.path.join(os.path.dirname(__file__), 'versions.cfg')
    config = ConfigParser.ConfigParser()
    config.optionxform = str
    config.readfp(open(versions_cfg))
    return dict(config.items('versions'))


def nailed_requires(requirements, pat=re.compile(r'^(.+)(\[.+\])?$')):
    """returns the requirements list with nailed versions"""
    versions = get_versions()
    res = []
    for req in requirements:
        if '[' in req:
            name = req.split('[', 1)[0]
        else:
            name = req
        if name in versions:
            res.append('%s==%s' % (req, versions[name]))
        else:
            res.append(req)
    return res


def read(path):
    return open(os.path.join(os.path.dirname(__file__), path)).read()


here = os.path.abspath(os.path.dirname(__file__))
readme = open(os.path.join(here, 'README.rst')).read()
changes = open(os.path.join(here, 'CHANGES.txt')).read()

requires = [
    'gevent',
    'pyramid',
    'validictory',
    'docutils',
]

test_requires = requires + [
    'webtest',
    'collective.xmltestreport',
    'requests',
    'sphinx',
]

setup(name='lovely.pyrest',
      version=version,
      description='rest framework for pyramids',
      long_description=readme + '\n\n' + changes,
      classifiers=[
          "Programming Language :: Python",
      ],
      author='lovely systems',
      author_email='office@lovelysystems.com',
      url='https://github.com/lovelysystems/lovely.pyrest',
      keywords='pyramid rest framework',
      license='apache license 2.0',
      packages=find_packages(),
      namespace_packages=['lovely'],
      include_package_data=True,
      extras_require=dict(
          test=nailed_requires(test_requires),
      ),
      zip_safe=False,
      install_requires=requires,
      tests_require=test_requires,
      test_suite="lovely.pyrest",
      )



import validictory
import copy


class ValidationException(Exception):
    """A validation exception"""


class Validator(validictory.SchemaValidator):
    """Extend the validator
    """

    def validate_oneOf(self, x, fieldname, schema, props=None):
        """Allows to provide multiple schemas for one property

        If the validation fails the error from the last validation is raised.
        """
        data = x[fieldname]
        error = None
        for s in props:
            try:
                self.validate(data, s)
                return
            except validictory.FieldValidationError as e:
                error = e
        if error:
            raise error


def validate(schema):
    """ A decorator to validate the kwargs of the decorated function against
    the given schema. The schema has to be a valid validictory schema
    """

    def wrapper(f):
        def validation_wrapper(*args, **kwargs):
            try:
                validictory.validate(kwargs, schema, Validator)
            except ValueError, error:
                raise ValidationException(error.message)
            return f(*args, **kwargs)
        # link doc with function docstring, so it's accessible for generating
        # documentation
        validation_wrapper.__doc__ = f.__doc__
        return validation_wrapper
    return wrapper


def non_required(schema):
    """ Helper function  to make all properties optional
    """
    s = copy.deepcopy(schema)
    for v in s["properties"].values():
        v['required'] = False
    return s
